/*
 * Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Radek Novacek <rnovacek@redhat.com>
 */

#include "connection.h"

#include <string.h>

#include "setting.h"

#include "connection_private.h"
#include "network_private.h"

Connection *connection_new(Network *network, const char *id, const char *name)
{
    Connection *connection = malloc(sizeof(Connection));
    connection->network = network;
    connection->id = (id != NULL) ? strdup(id) : NULL;
    connection->type = CONNECTION_TYPE_ETHERNET;
    connection->name = (name != NULL) ? strdup(name) : NULL;
    connection->uuid = NULL;
    connection->settings = NULL;
    connection->autoconnect = false;
    connection->port = NULL;
    connection->priv = connection_priv_new();
    return connection;
}

Connection *connection_clone(Connection *c)
{
    Connection *connection = connection_new(c->network, c->id, c->name);
    connection->type = c->type;
    connection->autoconnect = c->autoconnect;
    connection->port = c->port;
    for (size_t i = 0; i < settings_length(c->settings); ++i) {
        connection_add_setting(connection, setting_clone(settings_index(c->settings, i)));
    }
    return connection;
}

int connection_update(const Connection *connection, Connection *new_connection)
{
    return connection_priv_update(connection, new_connection);
}

const char *connection_get_uuid(const Connection *connection)
{
    return connection->uuid;
}

const char *connection_get_id(const Connection *connection)
{
    return connection->id;
}

const char *connection_get_name(const Connection *connection)
{
    return connection->name;
}

void connection_set_name(Connection *connection, const char *name)
{
    assert(name != NULL);
    if (connection->name != NULL) {
        free(connection->name);
    }
    connection->name = strdup(name);
}

ConnectionType connection_get_type(const Connection *connection)
{
    return connection->type;
}

void connection_set_type(Connection *connection, ConnectionType type)
{
    connection->type = type;
}

bool connection_get_autoconnect(const Connection *connection)
{
    return connection->autoconnect;
}

int connection_set_autoconnect(const Connection *connection, bool autoconnect)
{
    Connection new_connection;
    memcpy(&new_connection, connection, sizeof(Connection));
    new_connection.autoconnect = autoconnect;
    debug("Setting autoconnect to %s for connection %s",
          autoconnect ? "true" : "false", connection->name);
    int res = connection_update(connection, &new_connection);
    return res;
}

const Settings *connection_get_settings(const Connection *connection)
{
    return connection->settings;
}

const Port *connection_get_port(const Connection *connection)
{
    return connection->port;
}

void connection_add_setting(Connection *connection, Setting *setting)
{
    if (connection->settings == NULL) {
        connection->settings = settings_new(1);
    }
    settings_add(connection->settings, setting);
}

void connection_set_port(Connection *connection, Port *port)
{
    connection->port = port;
}

void connection_set_master_connection(Connection *connection, const Connection *master, SettingType type)
{
    connection_priv_set_master_connection(connection, master, type);
}

bool connection_compare(const Connection *c1, const Connection *c2)
{
    if (c1 == NULL || c2 == NULL) {
        return false;
    }
    if (c1->id == NULL || c2->id == NULL) {
        return false;
    }

    return strcmp(c1->id, c2->id) == 0;
}

int connection_delete(const Connection *connection)
{
    return connection_priv_delete(connection);
}

void connection_free(Connection *connection)
{
    if (connection->settings != NULL) {
        settings_free(connection->settings, true);
    }
    if (connection->uuid != NULL) {
        free(connection->uuid);
    }
    if (connection->id != NULL) {
        free(connection->id);
    }
    if (connection->name != NULL) {
        free(connection->name);
    }
    if (connection->priv != NULL) {
        connection_priv_free(connection->priv);
    }
    free(connection);
}

LIST_IMPL(Connection, connection)

Connection *connections_find_by_uuid(const Connections *connections, const char *uuid)
{
    if (uuid == NULL) {
        return NULL;
    }
    Connection *connection;
    for (size_t i = 0; i < connections_length(connections); ++i) {
        connection = connections_index(connections, i);
        if (connection->uuid != NULL && strcmp(connection->uuid, uuid) == 0) {
            return connection;
        }
    }
    return NULL;
}

Connection *connections_find_by_id(const Connections *connections, const char *id)
{
    if (id == NULL) {
        return NULL;
    }
    Connection *connection;
    for (size_t i = 0; i < connections_length(connections); ++i) {
        connection = connections_index(connections, i);
        if (connection->id != NULL && strcmp(connection->id, id) == 0) {
            return connection;
        }
    }
    return NULL;
}
