/*
 * Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Radek Novacek <rnovacek@redhat.com>
 */

#include "connection_nm.h"

#include <string.h>

#include "dbus_wrapper.h"
#include "errors.h"

#include "network_nm.h"
#include "nm_support.h"
#include "setting_nm.h"

#include "connection_private.h"
#include "port_private.h"
#include "network_private.h"
#include "setting_private.h"

const char *connection_type_to_string(ConnectionType type)
{
    switch (type) {
        case CONNECTION_TYPE_UNKNOWN:
            warn("connection_type_to_string called with type UNKNOWN");
            return NULL;
        case CONNECTION_TYPE_ETHERNET:
            return "802-3-ethernet";
        case CONNECTION_TYPE_BOND:
            return "bond";
        case CONNECTION_TYPE_BRIDGE:
            return "bridge";
    }
    error("connection_type_to_string called with undefined type %d", type);
    return NULL;
}

ConnectionType connection_type_from_string(const char *s)
{
    if (strcmp(s, "802-3-ethernet") == 0) {
        return CONNECTION_TYPE_ETHERNET;
    } else if (strcmp(s, "bond") == 0) {
        return CONNECTION_TYPE_BOND;
    } else if (strcmp(s, "bridge") == 0) {
        return CONNECTION_TYPE_BRIDGE;
    } else {
        warn("connection_type_from_string called with unknown string: %s", s);
        return CONNECTION_TYPE_UNKNOWN;
    }
}

typedef struct ConnectionPriv {
    DBusGProxy *proxy;
    char *master;
    char *slave_type;
} ConnectionPriv;

void connection_get_properties(Connection *connection);
void connection_updated_cb(void *proxy, Connection *connection);
void connection_removed_cb(void *proxy, Connection *connection);

Connection *connection_new_from_objectpath(Network *network, const char *objectpath)
{
    // Id and name will be filled later
    Connection *connection = connection_new(network, NULL, NULL);
    ConnectionPriv *priv = connection->priv;
    connection->uuid = strdup(objectpath);
    priv->proxy = dbus_g_proxy_new_for_name(network_priv_get_dbus_connection(network), NM_SERVICE_DBUS, objectpath, NM_DBUS_IFACE_SETTINGS_CONNECTION);

    connection_get_properties(connection);

    dbus_g_proxy_add_signal(priv->proxy, "Updated", G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(priv->proxy, "Updated", G_CALLBACK(connection_updated_cb), connection, NULL);

    dbus_g_proxy_add_signal(priv->proxy, "Removed", G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(priv->proxy, "Removed", G_CALLBACK(connection_removed_cb), connection, NULL);
    return connection;
}

void connection_read_properties(Connection *connection, GHashTable *hash)
{
    ConnectionPriv *priv = connection->priv;
    GHashTableIter iter;
    gpointer key, value;
    Setting *setting;
    const char *type, *s;
    GValue *v;
    char *id, *caption;
    g_hash_table_iter_init(&iter, hash);
    if (connection->settings != NULL) {
        settings_free(connection->settings, true);
    }
    connection->settings = settings_new(3);

    while (g_hash_table_iter_next(&iter, &key, &value)) {
        if (strcmp(key, "connection") == 0) {
            GHashTable *connection_map = value;
            if (!connection_map) {
                warn("Connection %s is invalid", connection->uuid);
                return;
            }
            if (connection->id != NULL) free(connection->id);
            connection->id = strdup(dbus_property_string(connection_map, "uuid"));
            if (connection->name != NULL) free(connection->name);
            connection->name = strdup(dbus_property_string(connection_map, "id"));
            GValue *v = g_hash_table_lookup(connection_map, "autoconnect");
            if (v) {
                connection->autoconnect = g_value_get_boolean(v);
            } else {
                connection->autoconnect = true;
            }
            type = dbus_property_string(connection_map, "type");
            if (type == NULL) {
                connection->type = CONNECTION_TYPE_UNKNOWN;
            } else {
                connection->type = connection_type_from_string(type);
            }

            v = g_hash_table_lookup(connection_map, "master");
            if (v) {
                s = g_value_get_string(v);
                if (s != NULL) {
                    if (priv->master != NULL) free(priv->master);
                    priv->master = strdup(s);
                }
            }
            v = g_hash_table_lookup(connection_map, "slave-type");
            if (v) {
                s = g_value_get_string(v);
                if (s != NULL) {
                    if (priv->slave_type != NULL) free(priv->slave_type);
                    priv->slave_type = strdup(s);
                }
            }
        } else if (strcmp(key, "802-3-ethernet") == 0) {
            v = g_hash_table_lookup((GHashTable *) value, "mac-address");
            if (v != NULL) {
                GByteArray *mac = g_value_get_boxed(v);
                char *mac_str = macFromGByteArray(mac);
                connection->port = network_port_by_mac(connection->network, mac_str);
                free(mac_str);
            }
        } else {
            setting = setting_from_hash(value, key);
            connection_add_setting(connection, setting);
        }
    }
    // Add ids and captions to the settings
    for (size_t i = 0; i < settings_length(connection->settings); ++i) {
        setting = settings_index(connection->settings, i);
        asprintf(&id, "%s_%ld", connection->id, i);
        setting->id = id;
        asprintf(&caption, "%s %ld", connection->name, i);
        setting->caption = caption;
    }
}

GHashTable *connection_to_hash(Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    GHashTable *hash = g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify) g_hash_table_destroy);
    GHashTable *subhash;
    Setting *setting;

    // connection
    subhash = g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify) g_value_free);
    g_hash_table_insert_string(subhash, "id", connection->name);

    if (connection->id == NULL) {
        connection->id = uuid_gen();
    }
    g_hash_table_insert_string(subhash, "uuid", connection->id);

    GValue *v = g_value_new(G_TYPE_BOOLEAN);
    g_value_set_boolean(v, connection->autoconnect);
    g_hash_table_insert(subhash, strdup("autoconnect"), v);

    g_hash_table_insert_string(subhash, "type", connection_type_to_string(connection->type));

    if (priv->master != NULL) {
        g_hash_table_insert_string(subhash, "master", priv->master);
    }
    if (priv->slave_type != NULL) {
        g_hash_table_insert_string(subhash, "slave-type", priv->slave_type);
    }

    g_hash_table_insert(hash, strdup("connection"), subhash);

    // connection type specific
    if (connection->type == CONNECTION_TYPE_ETHERNET) {
        subhash = g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify) g_value_free);
        if (connection->port != NULL) {
            g_hash_table_insert_boxed(subhash, "mac-address", DBUS_TYPE_G_UCHAR_ARRAY, macToGByteArray(connection->port->mac), true);
        }
        g_hash_table_insert(hash, strdup("802-3-ethernet"), subhash);
    }

    for (size_t i = 0; i < settings_length(connection->settings); ++i) {
        setting = settings_index(connection->settings, i);
        char *key;
        subhash = setting_to_hash(setting, &key);
        if (subhash != NULL) {
            g_hash_table_insert(hash, key, subhash);
        } else {
            free(key);
        }
    }
    g_hash_table_print(hash);
    return hash;
}

void *connection_priv_new(void)
{
    ConnectionPriv *p = malloc(sizeof(ConnectionPriv));
    p->proxy = NULL;
    p->master = NULL;
    p->slave_type = NULL;
    return p;
}

void connection_priv_free(void *priv)
{
    ConnectionPriv *p = priv;
    if (p->master != NULL) {
        free(p->master);
    }
    if (p->slave_type != NULL) {
        free(p->slave_type);
    }
    if (p->proxy != NULL) {
        g_object_unref(p->proxy);
    }
    free(p);
}

void connection_priv_set_master_connection(Connection *connection, const Connection *master, SettingType type)
{
    ConnectionPriv *priv = connection->priv;
    priv->master = strdup(master->id);
    if (type == SETTING_TYPE_BOND) {
        priv->slave_type = strdup("bond");
    }
    if (type == SETTING_TYPE_BRIDGE) {
        priv->slave_type = strdup("bridge");
    }
}

void connection_get_properties(Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    GHashTable *hash;
    GError *err = NULL;
    if (!dbus_g_proxy_call(priv->proxy, "GetSettings", &err, G_TYPE_INVALID,
        DBUS_TYPE_G_MAP_OF_MAP_OF_VARIANT, &hash, G_TYPE_INVALID)) {

        error("Call GetSetting of %s failed: %s", NM_DBUS_IFACE_SETTINGS_CONNECTION, err->message);
        return;
    }
    if (hash == NULL) {
        error("Connection %s doesn't have any settings", connection->uuid);
        return;
    }
    connection_read_properties(connection, hash);
}

void connection_updated_cb(void *proxy, Connection *connection)
{
    MUTEX_LOCK(connection->network);
    debug("Connection updated: %s (%s)", connection->id, connection->name);
    connection_get_properties(connection);
    MUTEX_UNLOCK(connection->network);
}

void connection_removed_cb(void *proxy, Connection *connection)
{
    assert(connection != NULL);
    Network *network = connection->network;
    MUTEX_LOCK(network);

    // NetworkManager sometimes sends Removed signal twice, we need to workaround it
    if (connection == NULL || strcmp(dbus_g_proxy_get_path(proxy), connection->uuid) != 0) {
        debug("Connection already deleted");
        MUTEX_UNLOCK(network);
        return;
    }
    debug("Connection deleted: %s (%s)", connection->id, connection->name);

    size_t i = 0;
    Connections *connections = connection->network->connections;
    for (i = 0; i < connections_length(connections); ++i) {
        if (strcmp(connections_index(connections, i)->uuid, connection->uuid) == 0) {
            break;
        }
    }
    if (i < connections_length(connections)) {
        connection_free(connections_pop(connections, i));
        connection = NULL;
    }
    MUTEX_UNLOCK(network);
}

int connection_priv_delete(const Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    GError *err = NULL;
    if (!dbus_g_proxy_call(priv->proxy, "Delete", &err, G_TYPE_INVALID, G_TYPE_INVALID)) {
        error("Deleting of connection failed: %s", err->message);
        return LMI_ERROR_CONNECTION_DELETE_FAILED;
    }
    return LMI_SUCCESS;
}

#ifndef TESTING
int connection_priv_update(const Connection *connection, Connection *new_connection)
{
    int res = LMI_SUCCESS;
    ConnectionPriv *priv = connection->priv;
    GError *err = NULL;
    GHashTable *new_hash = connection_to_hash(new_connection);

    if (!dbus_g_proxy_call(priv->proxy, "Update", &err, DBUS_TYPE_G_MAP_OF_MAP_OF_VARIANT, new_hash, G_TYPE_INVALID, G_TYPE_INVALID)) {
        error("Unable to update connection %s: %s",
              connection ? connection_get_name(connection) : "NULL",
              err->message);
        res = LMI_ERROR_CONNECTION_UPDATE_FAILED;
    }

    return res;
}
#endif
