/*
 * Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Radek Novacek <rnovacek@redhat.com>
 */

#include "globals.h"

#include <string.h>
#include <unistd.h>

#include <uuid/uuid.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/utsname.h>
#include <cmpimacs.h>

static char *_fqdn = NULL;
#ifdef DEBUG_ENABLED
static int _log_level = DEBUG;
#else
static int _log_level = WARN;
#endif

static const CMPIBroker *_cb = NULL;

#define LOG_ID "openlmi-networking"

char *getFQDN(void)
{
    struct utsname uts;
    if ((uname(&uts) > 0) && (strlen(uts.nodename) > 0)) {
        return strdup(uts.nodename);
    }
    char hostname[256];
    hostname[255] = '\0';
    if (gethostname(hostname, 255) == -1) {
        // FIXME: what to do, if we can't use gethostname?
        return NULL;
    }

    struct addrinfo hints;
    struct addrinfo *info = NULL, *p;
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC; // either IPV4 or IPV6
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;

    if (getaddrinfo(hostname, "http", &hints, &info) == 0) {
        for (p = info; p != NULL; p = p->ai_next) {
            if (p->ai_canonname && strstr(p->ai_canonname, "localhost") == NULL) {
                char *dn = strdup(p->ai_canonname);
                freeaddrinfo(info);
                return dn;
            }
        }
    }
    if (info != NULL) {
        freeaddrinfo(info);
    }
    return strdup(hostname);
}

const char *get_system_name()
{
    if (_fqdn == NULL) {
        _fqdn = getFQDN();
    }
    return _fqdn;
}

const char *get_system_creationg_class_name()
{
    return "Linux_ComputerSystem";
}

GValue *gvalue_new_from_value(GType type, gconstpointer data)
{
    GValue *v = g_new0(GValue, 1);
    g_value_init(v, type);
    g_value_take_boxed(v, data);
    return v;
}

char *ip4ToString(uint32_t i)
{
    struct in_addr in = { i };
    char *ip = malloc(INET_ADDRSTRLEN);
    if (inet_ntop(AF_INET, &in, ip, INET_ADDRSTRLEN) == NULL) {
        free(ip);
        return NULL;
    }
    return ip;
}

uint32_t ip4FromString(const char *ip)
{
    if (ip == NULL) {
        error("Invalid argument (null) for ip4FromString");
        return 0;
    }
    struct in_addr inp;
    if (inet_pton(AF_INET, ip, &inp) <= 0) {
        warn("IPv4 address %s is not valid.", ip);
        return 0;
    }
    return inp.s_addr;
}

char *ip6ToString(const struct in6_addr *i)
{
    if (i == NULL) {
        error("Invalid argument (null) for ip6ToString");
        return NULL;
    }
    char *ip = malloc(INET6_ADDRSTRLEN);
    if (ip == NULL) {
        error("Malloc failed");
        return NULL;
    }
    if (inet_ntop(AF_INET6, i, ip, INET6_ADDRSTRLEN) == NULL) {
        free(ip);
        return NULL;
    }
    return ip;

}

struct in6_addr *ip6FromString(const char *ip)
{
    if (ip == NULL) {
        error("Invalid argument (null) for ip6FromString");
        return NULL;
    }
    struct in6_addr *inp = malloc(sizeof(struct in6_addr));
    if (inp == NULL) {
        error("Malloc failed");
        return NULL;
    }
    if (inet_pton(AF_INET6, ip, inp) <= 0) {
        warn("IPv6 address %s is not valid.", ip);
        free(inp);
        return NULL;
    }
    return inp;
}

char *prefixToMask4(uint8_t prefix)
{
    return ip4ToString(((1 << prefix) - 1));
}

uint8_t netmaskToPrefix4(const char *netmask)
{
    if (netmask == NULL) {
        error("Invalid argument (null) for netmaskToPrefix4");
        return 0;
    }
    struct in_addr inp;
    if (inet_pton(AF_INET, netmask, &inp) <= 0) {
        warn("Invalid netmask: %s", netmask);
        return 0;
    }
    uint8_t ret = 0;
    while (inp.s_addr != 0) {
        ret += (inp.s_addr & 0x01);
        inp.s_addr >>= 1;
    }
    return ret;
}

char *macFromGByteArray(const GByteArray *mac)
{
    if (mac == NULL) {
        error("Invalid argument (null) for macFromGByteArray");
        return NULL;
    }
    char *m;
    asprintf(&m, "%02X:%02X:%02X:%02X:%02X:%02X",
             (int) mac->data[0], (int) mac->data[1],
             (int) mac->data[2], (int) mac->data[3],
             (int) mac->data[4], (int) mac->data[5]);
    return m;
}

GByteArray *macToGByteArray(const char *mac)
{
    if (mac == NULL) {
        error("Invalid argument (null) for macToGByteArray");
        return NULL;
    }
    unsigned int m[6];
    uint8_t mm[6];
    if (sscanf(mac, "%2X:%2X:%2X:%2X:%2X:%2X", &m[0], &m[1], &m[2], &m[3], &m[4], &m[5]) != 6) {
        error("MAC address is not valid: %s", mac);
        return NULL;
    }
    for (unsigned int i = 0; i < 6; ++i) {
        if (m[i] > 255) {
            error("MAC address is not valid: %s", mac);
            return NULL;
        }
        mm[i] = (uint8_t) m[i];
    }
    GByteArray *array = g_byte_array_sized_new(6);
    g_byte_array_append(array, mm, 6);
    return array;
}

char *uuid_gen(void)
{
    uuid_t uuid;
    char *buf = malloc(37 * sizeof(char));
    uuid_generate_random(uuid);
    uuid_unparse_lower(uuid, &buf[0]);
    return buf;
}

void init_logging(const CMPIBroker *cb)
{
    _cb = cb;
}

int log_level(void)
{
    return _log_level;
}

void set_log_level(int level)
{
    _log_level = level;
}

void _debug(int level, const char *file, int line, const char *format, ...)
{
    const char *lvl[] = { "NONE", "ERROR", "INFO", "WARNING", "DEBUG" };
    if (level > 4) {
        level = 4;
    }
    if (level < 1) {
        level = 1;
    }

    char *message, *text;
    va_list args;
    va_start(args, format);
    vasprintf(&message, format, args);
    va_end(args);
    asprintf(&text, "[%s] %s:%d\t%s", lvl[level], file, line, message);
    free(message);

    CMPIStatus rc;
    rc.rc = CMPI_RC_OK;
    if (_cb != NULL) {
        // try to use standard CMPI logging
        rc = _cb->eft->trace(_cb, CMPI_LEV_INFO, LOG_ID, text, NULL);
    }

#ifndef DEBUG_ENABLED
    if (_cb == NULL || rc.rc != CMPI_RC_OK) {
#else
    // Print to stderr too in DEBUG mode
    {
        (void) rc; // silence warning
#endif
        if (level > _log_level) {
            free(text);
            return;
        }

        fprintf(stderr, "%s", text);
        fprintf(stderr, "\n");
    }
    free(text);
}

GByteArray *ip6ArrayFromString(const char *addr)
{
    if (addr != NULL) {
        struct in6_addr *ip = ip6FromString(addr);
        if (ip == NULL) {
            return g_byte_array_new_take(calloc(sizeof(guint8), 16), 16);
        } else {
            GByteArray *array = g_byte_array_sized_new(16);
            g_byte_array_append(array, ip->s6_addr, 16);
            free(ip);
            return array;
        }
    } else {
        return g_byte_array_new_take(calloc(sizeof(guint8), 16), 16);
    }
}

char *ip6ArrayToString(GByteArray *array)
{
    char *ip = malloc(INET6_ADDRSTRLEN);
    if (inet_ntop(AF_INET6, array->data, ip, INET6_ADDRSTRLEN) == NULL) {
        free(ip);
        return NULL;
    }
    return ip;
}

GValue *g_value_new(GType t)
{
    GValue *v = calloc(1, sizeof(GValue));
    g_value_init(v, t);
    return v;
}

void g_value_free(GValue *v)
{
    g_value_unset(v);
    free(v);
}

void g_hash_table_insert_string(GHashTable *hash, const char *key, const char *value)
{
    GValue *v = g_value_new(G_TYPE_STRING);
    g_value_take_string(v, strdup(value));
    g_hash_table_insert(hash, strdup(key), v);
}

void g_hash_table_insert_boxed(GHashTable *hash, const char *key, GType t, gconstpointer value, bool take)
{
    GValue *v = g_value_new(t);
    if (take) {
        g_value_take_boxed(v, value);
    } else {
        g_value_set_boxed(v, value);
    }
    g_hash_table_insert(hash, strdup(key), v);
}

int find_index(const char *key, const char *list[])
{
    int i = 0;
    while (list[i] != NULL) {
        if (strcmp(key, list[i]) == 0) {
            return i;
        }
        i++;
    }
    // Return -1 if not found
    return -1;
}

char *id_to_instanceid(const char *id, const char *cls)
{
    assert(id != NULL);
    assert(cls != NULL);
    char *instanceid;
    asprintf(&instanceid, ORGID ":%s:%s", cls, id);
    return instanceid;
}


bool key_value_parse(char *input, char **key, char **value, char **saveptr)
{
    char *token = strtok_r(*saveptr != NULL ? NULL : input, ",", saveptr);
    if (token == NULL) {
        *key = NULL;
        *value = NULL;
        return false;
    }
    char *delim = strstr(token, "=");
    if (delim != NULL) {
        *delim = '\0';
        *value = delim + 1;
    } else {
        *value = NULL;
    }
    *key = token;
    return true;
}
