/*
 * Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Radek Novacek <rnovacek@redhat.com>
 */

#include "setting.h"

#include <string.h>

#include "port.h"

#include "setting_private.h"

Setting *setting_new(SettingType setting_type)
{
    Setting *setting = malloc(sizeof(Setting));
    setting->id = NULL;
    setting->caption = NULL;
    setting->type = setting_type;
    switch (setting_type) {
        case SETTING_TYPE_UNKNOWN:
            error("Cannot create setting with UNKNOWN type");
            break;
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            setting->typespec.ip.method = SETTING_METHOD_UNKNOWN;
            setting->typespec.ip.addresses = addresses_new(0);
            setting->typespec.ip.routes = routes_new(0);
            setting->typespec.ip.dns_servers = dns_servers_new(0);
            setting->typespec.ip.search_domains = search_domains_new(0);
            setting->typespec.ip.clientID = NULL;
            break;
        case SETTING_TYPE_BOND:
            setting->typespec.bond.interface_name = NULL;
            setting->typespec.bond.options = NULL;
            break;
        case SETTING_TYPE_BRIDGE:
            // TODO: add brigde options
            break;
        case SETTING_TYPE_WIRED:
            setting->typespec.wired.mac = NULL;
            break;
    }
    return setting;
}

Setting *setting_clone(const Setting *s)
{
    Setting *setting = setting_new(s->type);
    setting->id = strdup(s->id);
    setting->caption = strdup(s->caption);
    Address *address;
    size_t i;
    switch (s->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            setting->typespec.ip.method = s->typespec.ip.method;
            for (i = 0; i < addresses_length(s->typespec.ip.addresses); ++i) {
                address = addresses_index(s->typespec.ip.addresses, i);
                setting_add_ip_address(setting, s->typespec.ip.method,
                    address->addr, address->prefix, address->default_gateway);
            }
            for (i = 0; i < routes_length(s->typespec.ip.routes); ++i) {
                Route *route = routes_index(s->typespec.ip.routes, i);
                setting_add_route(setting, route->route, route->prefix, route->next_hop, route->metric);
            }
            for (i = 0; i < dns_servers_length(s->typespec.ip.dns_servers); ++i) {
                setting_add_dns_server(setting, dns_servers_index(s->typespec.ip.dns_servers, i)->server);
            }
            for (i = 0; i < search_domains_length(s->typespec.ip.search_domains); ++i) {
                setting_add_search_domain(setting, search_domains_index(s->typespec.ip.search_domains, i)->domain);
            }
            setting->typespec.ip.clientID = s->typespec.ip.clientID != NULL ? strdup(s->typespec.ip.clientID) : NULL;
            break;
        case SETTING_TYPE_BOND:
            setting->typespec.bond.interface_name =
                    s->typespec.bond.interface_name != NULL ?
                    strdup(s->typespec.bond.interface_name) :
                    NULL;
            setting->typespec.bond.options =
                    s->typespec.bond.options != NULL ?
                    strdup(setting->typespec.bond.options) :
                    NULL;
            break;
        case SETTING_TYPE_BRIDGE:
            // TODO: add brigde options
            break;
        case SETTING_TYPE_WIRED:
            setting->typespec.wired.mac =
                    s->typespec.wired.mac != NULL ?
                    strdup(s->typespec.wired.mac) :
                    NULL;
    }
    return setting;
}

void setting_set_options(Setting *setting, const char *options)
{
    char *key, *value, *saveptr = NULL, *opts = strdup(options);
    while (key_value_parse(opts, &key, &value, &saveptr)) {
        setting_set_option(setting, key, value);
    }
    free(opts);
}

const char *setting_get_option(Setting *setting, const char *key)
{
    switch (setting->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            if (strcasecmp(key, "clientID") == 0) {
                return setting->typespec.ip.clientID;
            }
            break;
        case SETTING_TYPE_BOND:
            if (strcasecmp(key, "interface-name") == 0) {
                return setting->typespec.bond.interface_name;
            }
            if (strcasecmp(key, "options") == 0) {
                return setting->typespec.bond.options;
            }
            break;
    }
    return NULL;
}

void setting_set_option(Setting *setting, const char *key, const char *value)
{
    switch (setting->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            if (strcasecmp(key, "clientID") == 0) {
                setting->typespec.ip.clientID = value != NULL ? strdup(value) : NULL;
            }
            break;
        case SETTING_TYPE_BOND:
            if (strcasecmp(key, "interface-name") == 0) {
                setting->typespec.bond.interface_name = value != NULL ? strdup(value) : NULL;
            }
            if (strcasecmp(key, "options") == 0) {
                setting->typespec.bond.options = value != NULL ? strdup(value) : NULL;
            }
            break;
    }
}

void setting_clear_addresses(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    while (addresses_length(setting->typespec.ip.addresses) > 0) {
        address_free(addresses_pop(setting->typespec.ip.addresses, 0));
    }
}

void setting_add_ip_address(Setting *setting, SettingMethod method,
                              const char *address, uint8_t prefix,
                              const char *default_gateway)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);

    setting->typespec.ip.method = method;
    Address *addr = address_new(setting->type == SETTING_TYPE_IPv4 ? IPv4 : IPv6);
    addr->addr = strdup(address);
    addr->prefix = prefix;
    if (default_gateway != NULL) {
        addr->default_gateway = strdup(default_gateway);
    }
    addresses_add(setting->typespec.ip.addresses, addr);
}

SettingType setting_get_type(const Setting *setting)
{
    return setting->type;
}

const char *setting_get_id(const Setting *setting)
{
    return setting->id;
}

const char *setting_get_caption(const Setting *setting)
{
    return setting->caption;
}

Addresses *setting_get_addresses(const Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return setting->typespec.ip.addresses;
}

SettingMethod setting_get_method(const Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return setting->typespec.ip.method;
}


void setting_set_method(Setting *setting, SettingMethod method)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    setting->typespec.ip.method = method;
}

const char *setting_get_clientID(const Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return setting->typespec.ip.clientID;
}

void setting_set_clientID(Setting *setting, const char *clientID)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    setting->typespec.ip.clientID = strdup(clientID);
}

void setting_clear_dns_servers(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    while (dns_servers_length(setting->typespec.ip.dns_servers) > 0) {
        dns_server_free(dns_servers_pop(setting->typespec.ip.dns_servers, 0));
    }
}

void setting_add_dns_server(Setting *setting, const char *dns_server)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    dns_servers_add(setting->typespec.ip.dns_servers,
                    dns_server_new(setting->type == SETTING_TYPE_IPv4 ? IPv4 : IPv6, dns_server));
}

size_t setting_get_dns_servers(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return dns_servers_length(setting->typespec.ip.dns_servers);
}

size_t setting_get_dns_servers_length(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return dns_servers_length(setting->typespec.ip.dns_servers);
}

const char *setting_get_dns_server(Setting *setting, size_t index)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    assert(index < dns_servers_length(setting->typespec.ip.dns_servers));
    return dns_servers_index(setting->typespec.ip.dns_servers, index)->server;
}

void setting_clear_search_domains(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    while (search_domains_length(setting->typespec.ip.search_domains) > 0) {
        search_domain_free(search_domains_pop(setting->typespec.ip.search_domains, 0));
    }
}

void setting_add_search_domain(Setting *setting, const char *search_domain)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    search_domains_add(setting->typespec.ip.search_domains,
                       search_domain_new(setting->type == SETTING_TYPE_IPv4 ? IPv4 : IPv6, search_domain));
}

size_t setting_get_search_domains_length(Setting *setting)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    return search_domains_length(setting->typespec.ip.search_domains);
}

const char *setting_get_search_domain (Setting *setting, size_t index)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    assert(index < search_domains_length(setting->typespec.ip.search_domains));
    return search_domains_index(setting->typespec.ip.search_domains, index)->domain;
}

void setting_add_route(Setting *setting, const char *route, uint32_t prefix, const char *next_hop, uint32_t metric)
{
    assert(setting->type == SETTING_TYPE_IPv4 || setting->type == SETTING_TYPE_IPv6);
    Route *r = route_new(setting->type == SETTING_TYPE_IPv4 ? IPv4 : IPv6);
    r->route = route != NULL ? strdup(route) : NULL;
    r->prefix = prefix;
    r->next_hop = next_hop != NULL ? strdup(next_hop) : NULL;
    r->metric = metric;
    routes_add(setting->typespec.ip.routes, r);
}

void setting_free(Setting *setting)
{
    if (setting->id != NULL)
        free(setting->id);
    if (setting->caption != NULL)
        free(setting->caption);
    switch (setting->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            addresses_free(setting->typespec.ip.addresses, true);
            routes_free(setting->typespec.ip.routes, true);
            dns_servers_free(setting->typespec.ip.dns_servers, true);
            search_domains_free(setting->typespec.ip.search_domains, true);
            if (setting->typespec.ip.clientID != NULL) {
                free(setting->typespec.ip.clientID);
            }
            break;
        case SETTING_TYPE_BOND:
            if (setting->typespec.bond.interface_name != NULL) {
                free(setting->typespec.bond.interface_name);
            }
            if (setting->typespec.bond.options != NULL) {
                free(setting->typespec.bond.options);
            }
            break;
        case SETTING_TYPE_WIRED:
            if (setting->typespec.wired.mac != NULL) {
                free(setting->typespec.wired.mac);
            }
            break;
        case SETTING_TYPE_UNKNOWN:
            break;
    }
    free(setting);
}

LIST_IMPL(Setting, setting)

Setting *settings_find_by_type(const Settings *settings, SettingType type)
{
    if (settings == NULL) {
        return NULL;
    }
    for (size_t i = 0; i < settings->length; ++i) {
        if (settings->settings[i]->type == type) {
            return settings->settings[i];
        }
    }
    return NULL;
}
