/*
 * Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Radek Novacek <rnovacek@redhat.com>
 */

#include "setting_nm.h"

#include <string.h>

#include "connection.h"

#include "dbus_wrapper.h"
#include "nm_support.h"
#include "setting_private.h"
#include "connection_private.h"

/* Must have same order as SettingType in globals.h */
const char *setting_type_strings[] = {
    "ipv4",
    "ipv6",
    "wired",
    "bond",
    "bridge",
    NULL
};

GHashTable *setting_to_hash(Setting *setting, char **key)
{
    GHashTable *hash = g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify) g_value_free);

    *key = strdup(setting_type_strings[setting->type]);
    size_t i, j;
    switch (setting->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            switch (setting->typespec.ip.method) {
                case SETTING_METHOD_DHCP:
                case SETTING_METHOD_STATELESS:
                    g_hash_table_insert_string(hash, "method", "auto");
                    break;
                case SETTING_METHOD_DHCPv6:
                    g_hash_table_insert_string(hash, "method", "dhcp");
                    break;
                case SETTING_METHOD_LINK_LOCAL:
                    g_hash_table_insert_string(hash, "method", "link-local");
                    break;
                case SETTING_METHOD_DISABLED:
                    if (setting->type == SETTING_TYPE_IPv4) {
                        g_hash_table_insert_string(hash, "method", "disabled");
                    } else {
                        g_hash_table_insert_string(hash, "method", "ignore");
                    }
                    break;
                case SETTING_METHOD_STATIC:
                    g_hash_table_insert_string(hash, "method", "manual");
                    if (setting->type == SETTING_TYPE_IPv4) {
                        GPtrArray *ip4addresses = g_ptr_array_new_with_free_func((GDestroyNotify) g_array_free);
                        for (i = 0; i < addresses_length(setting->typespec.ip.addresses); ++i) {
                            g_ptr_array_add(ip4addresses,
                                    g_array_ref(
                                        address_to_ipv4_array(
                                            addresses_index(setting->typespec.ip.addresses, i))));
                        }
                        g_hash_table_insert_boxed(hash, "addresses", NM_TYPE_IP4_ADDRESSES, ip4addresses, true);

                        if (setting->typespec.ip.dns_servers != NULL) {
                            DNSServer *dns_server;
                            GArray *dns = g_array_new(true, true, sizeof(guint32));
                            for (j = 0; j < dns_servers_length(setting->typespec.ip.dns_servers); ++j) {
                                dns_server = dns_servers_index(setting->typespec.ip.dns_servers, j);
                                if (dns_server->type == IPv4) {
                                    guint32 ip = ip4FromString(dns_server->server);
                                    g_array_append_val(dns, ip);
                                }
                            }
                            g_hash_table_insert_boxed(hash, "dns", NM_TYPE_DNS4_SERVERS, dns, true);

                        }
                        if (setting->typespec.ip.routes != NULL) {
                            GPtrArray *routes = g_ptr_array_new_with_free_func((GDestroyNotify) g_array_free);
                            GArray *r;
                            guint32 ip;
                            for (j = 0; j < routes_length(setting->typespec.ip.routes); ++j) {
                                Route *route = routes_index(setting->typespec.ip.routes, j);
                                r = g_array_new(true, true, sizeof(guint32));
                                ip = ip4FromString(route->route);
                                g_array_append_val(r, ip);
                                g_array_append_val(r, route->prefix);
                                ip = ip4FromString(route->next_hop);
                                g_array_append_val(r, ip);
                                g_array_append_val(r, route->metric);
                                g_array_ref(r);
                                g_ptr_array_add(routes, r);
                            }
                            g_hash_table_insert_boxed(hash, "routes", NM_TYPE_IP4_ROUTES, routes, true);
                        }
                    } else {
                        GPtrArray *ip6addresses = g_ptr_array_new_with_free_func((GDestroyNotify) g_value_array_free);
                        for (i = 0; i < addresses_length(setting->typespec.ip.addresses); ++i) {
                            g_ptr_array_add(ip6addresses, address_to_ipv6_array(addresses_index(setting->typespec.ip.addresses, i)));
                        }
                        g_hash_table_insert_boxed(hash, "addresses", NM_TYPE_IP6_ADDRESSES, ip6addresses, false);

                        if (setting->typespec.ip.dns_servers != NULL) {
                            GPtrArray *dns = g_ptr_array_new_with_free_func((GDestroyNotify) g_byte_array_free);
                            DNSServer *dns_server;
                            for (j = 0; j < dns_servers_length(setting->typespec.ip.dns_servers); ++j) {
                                dns_server = dns_servers_index(setting->typespec.ip.dns_servers, j);
                                if (dns_server->type == IPv6) {
                                    GByteArray *ip = ip6ArrayFromString(dns_server->server);
                                    g_byte_array_ref(ip);
                                    g_ptr_array_add(dns, ip);
                                }
                            }
                            g_hash_table_insert_boxed(hash, "dns", NM_TYPE_DNS6_SERVERS, dns, true);
                        }
                        if (setting->typespec.ip.routes != NULL) {
                            GPtrArray *routes = g_ptr_array_new_with_free_func((GDestroyNotify) g_value_array_free);
                            GValueArray *array;
                            for (j = 0; j < routes_length(setting->typespec.ip.routes); ++j) {
                                Route *route = routes_index(setting->typespec.ip.routes, j);
                                array = g_value_array_new(4);
                                GValue v = G_VALUE_INIT;

                                // route
                                g_value_init(&v, DBUS_TYPE_G_UCHAR_ARRAY);
                                g_value_take_boxed(&v, ip6ArrayFromString(route->route));
                                g_value_array_append(array, &v);
                                g_value_unset(&v);
                                // prefix
                                g_value_init(&v, G_TYPE_UINT);
                                g_value_set_uint(&v, route->prefix);
                                g_value_array_append(array, &v);
                                g_value_unset(&v);
                                // next_hop
                                g_value_init(&v, DBUS_TYPE_G_UCHAR_ARRAY);
                                g_value_take_boxed(&v, ip6ArrayFromString(route->next_hop));
                                g_value_array_append(array, &v);
                                g_value_unset(&v);
                                // metric
                                g_value_init(&v, G_TYPE_UINT);
                                g_value_set_uint(&v, route->metric);
                                g_value_array_append(array, &v);
                                g_value_unset(&v);

                                g_ptr_array_add(routes, array);
                            }
                            g_hash_table_insert_boxed(hash, "routes", NM_TYPE_IP6_ROUTES, routes, false);
                        }
                    }
                    break;
                case SETTING_METHOD_UNKNOWN:
                    error("Cannot convert UNKNOWN setting to hash");
                    return NULL;
            }
            if (setting->typespec.ip.search_domains != NULL && search_domains_length(setting->typespec.ip.search_domains) > 0) {
                size_t len = search_domains_length(setting->typespec.ip.search_domains);
                char **search_domains = malloc((len + 1) * sizeof(char *));
                for (size_t j = 0; j < len; ++j) {
                    search_domains[j] = strdup(search_domains_index(setting->typespec.ip.search_domains, j)->domain);
                }
                search_domains[len] = NULL;
                g_hash_table_insert_boxed(hash, "dns-search", NM_TYPE_SEARCH_DOMAINS, search_domains, true);
            }
            break;
        case SETTING_TYPE_WIRED:
            break;
        case SETTING_TYPE_BOND:
            if (setting->typespec.bond.interface_name) {
                g_hash_table_insert_string(hash, "interface-name", setting->typespec.bond.interface_name);
            }
            if (setting->typespec.bond.options) {
                char *opts = strdup(setting->typespec.bond.options), *key, *value, *saveptr = NULL;
                GHashTable *option_hash = g_hash_table_new(g_str_hash, g_str_equal);
                while (key_value_parse(opts, &key, &value, &saveptr)) {
                    g_hash_table_insert_string(option_hash, key, value);
                }
                g_hash_table_insert_boxed(hash, "options", DBUS_TYPE_G_MAP_OF_STRING, option_hash, true);
            }
            break;
        default:
            error("Can't create setting with unknown type");
    }
    return hash;
}

GValue *g_hash_table_lookup_check(GHashTable *hash, const char *key, GType type)
{
    GValue *v = g_hash_table_lookup(hash, key);
    if (v == NULL) {
        return NULL;
    }
    if (!G_VALUE_HOLDS(v, type)) {
        error("Wrong type for key \"%s\": %s", key, G_VALUE_TYPE_NAME(v));
        return NULL;
    }
    return v;
}

Setting *setting_from_hash(GHashTable *hash, const char *key)
{
    Setting *setting = setting_new(find_index(key, setting_type_strings));
    const char *str;
    GValue *value;
    switch (setting->type) {
        case SETTING_TYPE_IPv4:
        case SETTING_TYPE_IPv6:
            value = g_hash_table_lookup_check(hash, "method", G_TYPE_STRING);
            if (value) {
                str = g_value_get_string(value);
                if (str) {
                    if (strcmp(str, "auto") == 0) {
                        if (setting->type == SETTING_TYPE_IPv4) {
                            setting->typespec.ip.method = SETTING_METHOD_DHCP;
                        } else {
                            setting->typespec.ip.method = SETTING_METHOD_STATELESS;
                        }
                    } else if (strcmp(str, "dhcp") == 0) {
                        // DHCP option is only valid for IPv6
                        setting->typespec.ip.method = SETTING_METHOD_DHCPv6;
                    } else if (strcmp(str, "link-local") == 0) {
                        setting->typespec.ip.method = SETTING_METHOD_LINK_LOCAL;
                    } else if (strcmp(str, "manual") == 0) {
                        setting->typespec.ip.method = SETTING_METHOD_STATIC;
                    } else if ((strcmp(str, "ignore") == 0) || (strcmp(str, "disabled") == 0)) {
                        setting->typespec.ip.method = SETTING_METHOD_DISABLED;
                        return setting;
                    } else {
                        warn("Unknown method: %s", str);
                    }
                }
            }
            if (setting->type == SETTING_TYPE_IPv4) {
                value = g_hash_table_lookup_check(hash, "addresses", NM_TYPE_IP4_ADDRESSES);
            } else {
                value = g_hash_table_lookup_check(hash, "addresses", NM_TYPE_IP6_ADDRESSES);
            }
            if (value) {
                GPtrArray *addresses = g_value_get_boxed(value);
                Address *address;
                for (guint i = 0; i < addresses->len; ++i) {
                    if (setting->type == SETTING_TYPE_IPv4) {
                        address = ipv4_array_to_address(g_ptr_array_index(addresses, i));
                    } else {
                        address = ipv6_array_to_address(g_ptr_array_index(addresses, i));
                    }
                    if (address != NULL) {
                        addresses_add(setting->typespec.ip.addresses, address);
                    }
                }
            }
            if (setting->type == SETTING_TYPE_IPv4) {
                value = g_hash_table_lookup_check(hash, "routes", NM_TYPE_IP4_ROUTES);
            } else {
                value = g_hash_table_lookup_check(hash, "routes", NM_TYPE_IP6_ROUTES);
            }
            if (value) {
                GPtrArray *nm_routes = g_value_get_boxed(value);
                Route *route;
                if (setting->type == SETTING_TYPE_IPv4) {
                    GArray *nm_route;
                    for (guint i = 0; i < nm_routes->len; ++i) {
                        nm_route = g_ptr_array_index(nm_routes, i);
                        route = route_new(IPv4);
                        route->route = ip4ToString(g_array_index(nm_route, guint32, 0));
                        route->prefix = g_array_index(nm_route, guint32, 1);
                        route->next_hop = ip4ToString(g_array_index(nm_route, guint32, 2));
                        route->metric = g_array_index(nm_route, guint32, 3);
                        routes_add(setting->typespec.ip.routes, route);
                    }
                } else {
                    GValueArray *nm_route;
                    for (guint i = 0; i < nm_routes->len; ++i) {
                        nm_route = g_ptr_array_index(nm_routes, i);
                        route = route_new(IPv6);
                        route->route = ip6ArrayToString(g_value_get_boxed(g_value_array_get_nth(nm_route, 0)));
                        route->prefix = g_value_get_uint(g_value_array_get_nth(nm_route, 1));
                        route->next_hop = ip6ArrayToString(g_value_get_boxed(g_value_array_get_nth(nm_route, 2)));
                        route->metric = g_value_get_uint(g_value_array_get_nth(nm_route, 3));
                        routes_add(setting->typespec.ip.routes, route);
                    }
                }
            }

            if (setting->type == SETTING_TYPE_IPv4) {
                value = g_hash_table_lookup_check(hash, "dns", NM_TYPE_DNS4_SERVERS);
            } else {
                value = g_hash_table_lookup_check(hash, "dns", NM_TYPE_DNS6_SERVERS);
            }
            if (value) {
                if (setting->type == SETTING_TYPE_IPv4) {
                    GArray *dns_servers = g_value_get_boxed(value);
                    for (guint i = 0; i < dns_servers->len; ++i) {
                        char *ip = ip4ToString(g_array_index(dns_servers, guint32, i));
                        dns_servers_add(setting->typespec.ip.dns_servers, dns_server_new(IPv4, ip));
                        free(ip);
                    }
                } else {
                    GPtrArray *dns_servers = g_value_get_boxed(value);
                    for (guint i = 0; i < dns_servers->len; ++i) {
                        char *ip = ip6ArrayToString(g_ptr_array_index(dns_servers, i));
                        dns_servers_add(setting->typespec.ip.dns_servers, dns_server_new(IPv6, ip));
                        free(ip);
                    }
                }
            }
            value = g_hash_table_lookup_check(hash, "dns-search", NM_TYPE_SEARCH_DOMAINS);
            if (value) {
                char **search_domain_ptr, **search_domains = g_value_get_boxed(value);
                for (search_domain_ptr = search_domains; *search_domain_ptr; search_domain_ptr++) {
                    search_domains_add(setting->typespec.ip.search_domains,
                            search_domain_new(setting->type == SETTING_TYPE_IPv4 ? IPv4 : IPv6,
                                              *search_domain_ptr));
                }
            }

            break;
        case SETTING_TYPE_WIRED:
            str = g_hash_table_lookup(hash, "mac-address");
            if (str) {
                setting->typespec.wired.mac = strdup(str);
            }
            break;
        case SETTING_TYPE_BOND:
            str = g_hash_table_lookup(hash, "interface-name");
            if (str) {
                setting->typespec.bond.interface_name = strdup(str);
            }
            value = g_hash_table_lookup(hash, "options");
            if (value) {
                GHashTable *option_hash = g_value_get_boxed(value);
                GHashTableIter iter;
                g_hash_table_iter_init(&iter, option_hash);
                char *subkey, *value;
                char *options = NULL;
                gpointer pkey, pvalue;
                while (g_hash_table_iter_next(&iter, &pkey, &pvalue)) {
                    subkey = (char *) pkey;
                    value = (char *) pvalue;
                    if (options == NULL) {
                        asprintf(&options, "%s=%s", subkey, value);
                    } else {
                        asprintf(&options, "%s,%s=%s", options, subkey, value);
                    }
                }
                if (options) {
                    debug("Bond options: %s", options);
                    setting->typespec.bond.options = options;
                }
            }
            break;
        case SETTING_TYPE_UNKNOWN:
            error("Can't handle setting (%s) with UNKNOWN", key);
            setting_free(setting);
            return NULL;
            break;
    }
    return setting;
}

/*
const char *setting_get_nm_method(Setting *setting)
{
    switch (setting->type) {
        case SETTING_STATIC:
            return (setting->protocol_type == IPv4) ?
                    NM_SETTING_IP4_CONFIG_METHOD_MANUAL :
                    NM_SETTING_IP6_CONFIG_METHOD_MANUAL;
            break;
        case SETTING_DHCP:
            return (setting->protocol_type == IPv4) ?
                    NM_SETTING_IP4_CONFIG_METHOD_AUTO :
                    NM_SETTING_IP6_CONFIG_METHOD_AUTO;
            break;
        case SETTING_LINK_LOCAL:
            return (setting->protocol_type == IPv4) ?
                    NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL :
                    NM_SETTING_IP6_CONFIG_METHOD_LINK_LOCAL;
            break;
        default:
            return (setting->protocol_type == IPv4) ?
                    NM_SETTING_IP4_CONFIG_METHOD_DISABLED :
                    NM_SETTING_IP6_CONFIG_METHOD_IGNORE;
    }
}
*/
